{ File   : HPFExt.dpr
  Author : Deniz Oezmen
  Created: 2007-01-14
  Changed: 2007-01-14

  The Last Express HPF extractor.
}
program HPFExt;

{$APPTYPE CONSOLE}

uses
  SysUtils;

type
  THPFHeader = packed record
                 NumFiles: Longword;
               end;

  THPFEntry  = packed record
                 Name      : array[0..11] of Char;
                 Sector,
                 NumSectors: Longword;
                 Unknown   : Word;
               end;

var
  HPFHeader: THPFHeader;
  HPFFile  : file;

function OpenHPFFile(var HPFFile: file; HPFName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(HPFFile, HPFName);

  {$I-}
  Reset(HPFFile, 1);
  {$I+}
  OpenHPFFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(HPFFile, HPFHeader, SizeOf(THPFHeader), BytesRead);
  ReadHeader := (BytesRead = SizeOf(THPFHeader))
end;

procedure ExtractFiles;
var
  i       : Longint;
  FileList: array of THPFEntry;
  Buffer  : array of Byte;
  OutFile : file;
begin
  SetLength(FileList, HPFHeader.NumFiles);

  {$I-}
  BlockRead(HPFFile, FileList[0], SizeOf(THPFEntry) * HPFHeader.NumFiles);
  {$I+}

  if IOResult <> 0 then
  begin
    WriteLn('  Error reading file table.');
    Finalize(FileList);
    Exit
  end;

  for i := 0 to High(FileList) do
    with FileList[i] do
    begin
      WriteLn(Format('  %s @ sector %d up by %d sector(s)',
        [string(Name), Sector, NumSectors]));

      FileMode := fmOpenWrite;
      AssignFile(OutFile, Name);
      {$I-}
      Rewrite(OutFile, 1);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    Error creating output file.');
        Continue
      end;

      SetLength(Buffer, NumSectors * 2048);

      {$I-}
      Seek(HPFFile, Sector * 2048);
      BlockRead(HPFFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('    Error reading from input file.')
      else
      begin
        {$I-}
        BlockWrite(OutFile, Buffer[0], Length(Buffer));
        {$I+}
        if IOResult <> 0 then
          WriteLn('    Error writing to output file.')
      end        
    end;

  Finalize(FileList);
  Finalize(Buffer)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  HPFName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      HPFName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenHPFFile(HPFFile, HPFName) then
        if ReadHeader then
        begin
          Writeln('Extracting from ', HPFName, ' ...');
          ExtractFiles;
          Close(HPFFile)
        end
        else
          Writeln(HPFName, ' is not a valid HPF file.')
      else
        Writeln(HPFName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

